/******************************************************************************
  Copyright 2013 Tom Cornall

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.  
 ******************************************************************************/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define KBD_COLS 23
#define KBD_ROWS 4

int hflag;
int lflag;
FILE *inputFile;
FILE *outputFile;
unsigned char matrix[KBD_COLS][KBD_ROWS];

typedef struct
{
    const char *tok;
    unsigned char code;
} ScanCode;

ScanCode scanCodes[] = {
    { .tok = "ignore",            .code = 0x00 },
    { .tok = "CAL_DN",            .code = 0x01 },
    { .tok = "CAL_UP",            .code = 0x02 },
    { .tok = "A",                 .code = 0x04 },
    { .tok = "B",                 .code = 0x05 },
    { .tok = "C",                 .code = 0x06 },
    { .tok = "D",                 .code = 0x07 },
    { .tok = "E",                 .code = 0x08 },
    { .tok = "F",                 .code = 0x09 },
    { .tok = "G",                 .code = 0x0a },
    { .tok = "H",                 .code = 0x0b },
    { .tok = "I",                 .code = 0x0c },
    { .tok = "J",                 .code = 0x0d },
    { .tok = "K",                 .code = 0x0e },
    { .tok = "L",                 .code = 0x0f },
    { .tok = "M",                 .code = 0x10 },
    { .tok = "N",                 .code = 0x11 },
    { .tok = "O",                 .code = 0x12 },
    { .tok = "P",                 .code = 0x13 },
    { .tok = "Q",                 .code = 0x14 },
    { .tok = "R",                 .code = 0x15 },
    { .tok = "S",                 .code = 0x16 },
    { .tok = "T",                 .code = 0x17 },
    { .tok = "U",                 .code = 0x18 },
    { .tok = "V",                 .code = 0x19 },
    { .tok = "W",                 .code = 0x1a },
    { .tok = "X",                 .code = 0x1b },
    { .tok = "Y",                 .code = 0x1c },
    { .tok = "Z",                 .code = 0x1d },
    { .tok = "1",                 .code = 0x1e },
    { .tok = "2",                 .code = 0x1f },
    { .tok = "3",                 .code = 0x20 },
    { .tok = "4",                 .code = 0x21 },
    { .tok = "5",                 .code = 0x22 },
    { .tok = "6",                 .code = 0x23 },
    { .tok = "7",                 .code = 0x24 },
    { .tok = "8",                 .code = 0x25 },
    { .tok = "9",                 .code = 0x26 },
    { .tok = "0",                 .code = 0x27 },
    { .tok = "Enter",             .code = 0x28 },
    { .tok = "Escape",            .code = 0x29 },
    { .tok = "Backspace",         .code = 0x2a },
    { .tok = "Tab",               .code = 0x2b },
    { .tok = "Space",             .code = 0x2c },
    { .tok = "-",                 .code = 0x2d },
    { .tok = "=",                 .code = 0x2e },
    { .tok = "[",                 .code = 0x2f },
    { .tok = "]",                 .code = 0x30 },
    { .tok = "\\",                .code = 0x31 },
    { .tok = "NonUS#",            .code = 0x32 },
    { .tok = ";",                 .code = 0x33 },
    { .tok = "'",                 .code = 0x34 },
    { .tok = "`",                 .code = 0x35 },
    { .tok = ",",                 .code = 0x36 },
    { .tok = ".",                 .code = 0x37 },
    { .tok = "/",                 .code = 0x38 },
    { .tok = "Caps",              .code = 0x39 },
    { .tok = "F1",                .code = 0x3a },
    { .tok = "F2",                .code = 0x3b },
    { .tok = "F3",                .code = 0x3c },
    { .tok = "F4",                .code = 0x3d },
    { .tok = "F5",                .code = 0x3e },
    { .tok = "F6",                .code = 0x3f },
    { .tok = "F7",                .code = 0x40 },
    { .tok = "F8",                .code = 0x41 },
    { .tok = "F9",                .code = 0x42 },
    { .tok = "F10",               .code = 0x43 },
    { .tok = "F11",               .code = 0x44 },
    { .tok = "F12",               .code = 0x45 },
    { .tok = "PrintScr",          .code = 0x46 },
    { .tok = "ScrollLock",        .code = 0x47 },
    { .tok = "Pause",             .code = 0x48 },
    { .tok = "Insert",            .code = 0x49 },
    { .tok = "Home",              .code = 0x4a },
    { .tok = "PageUp",            .code = 0x4b },
    { .tok = "Del",               .code = 0x4c },
    { .tok = "End",               .code = 0x4d },
    { .tok = "PageDown",          .code = 0x4e },
    { .tok = "RightArrow",        .code = 0x4f },
    { .tok = "LeftArrow",         .code = 0x50 },
    { .tok = "DownArrow",         .code = 0x51 },
    { .tok = "UpArrow",           .code = 0x52 },
    { .tok = "NumLock",           .code = 0x53 },
    { .tok = "KP/",               .code = 0x54 },
    { .tok = "KP*",               .code = 0x55 },
    { .tok = "KP-",               .code = 0x56 },
    { .tok = "KP+",               .code = 0x57 },
    { .tok = "KPEnter",           .code = 0x58 },
    { .tok = "KP1",               .code = 0x59 },
    { .tok = "KP2",               .code = 0x5a },
    { .tok = "KP3",               .code = 0x5b },
    { .tok = "KP4",               .code = 0x5c },
    { .tok = "KP5",               .code = 0x5d },
    { .tok = "KP6",               .code = 0x5e },
    { .tok = "KP7",               .code = 0x5f },
    { .tok = "KP8",               .code = 0x60 },
    { .tok = "KP9",               .code = 0x61 },
    { .tok = "KP0",               .code = 0x62 },
    { .tok = "KP.",               .code = 0x63 },
    { .tok = "NonUS\\",           .code = 0x64 },
    { .tok = "App",               .code = 0x65 },
    { .tok = "Power",             .code = 0x66 },
    { .tok = "KP=",               .code = 0x67 },
    { .tok = "F13",               .code = 0x68 },
    { .tok = "F14",               .code = 0x69 },
    { .tok = "F15",               .code = 0x6a },
    { .tok = "F16",               .code = 0x6b },
    { .tok = "F17",               .code = 0x6c },
    { .tok = "F18",               .code = 0x6d },
    { .tok = "F19",               .code = 0x6e },
    { .tok = "F20",               .code = 0x6f },
    { .tok = "F21",               .code = 0x70 },
    { .tok = "F22",               .code = 0x71 },
    { .tok = "F23",               .code = 0x72 },
    { .tok = "F24",               .code = 0x73 },
    { .tok = "Execute",           .code = 0x74 },
    { .tok = "Help",              .code = 0x75 },
    { .tok = "Menu",              .code = 0x76 },
    { .tok = "Select",            .code = 0x77 },
    { .tok = "Stop",              .code = 0x78 },
    { .tok = "Again",             .code = 0x79 },
    { .tok = "Undo",              .code = 0x7a },
    { .tok = "Cut",               .code = 0x7b },
    { .tok = "Copy",              .code = 0x7c },
    { .tok = "Paste",             .code = 0x7d },
    { .tok = "Find",              .code = 0x7e },
    { .tok = "Mute",              .code = 0x7f },
    { .tok = "VolUp",             .code = 0x80 },
    { .tok = "VolDown",           .code = 0x81 },
    { .tok = "LockingCaps",       .code = 0x82 },
    { .tok = "LockingNumLock",    .code = 0x83 },
    { .tok = "LockingScrollLock", .code = 0x84 },
    { .tok = "KP,",               .code = 0x85 },
    { .tok = "KP=",               .code = 0x86 },
    { .tok = "Intl1",             .code = 0x87 },
    { .tok = "Intl2",             .code = 0x88 },
    { .tok = "Intl3",             .code = 0x89 },
    { .tok = "Intl4",             .code = 0x8a },
    { .tok = "Intl5",             .code = 0x8b },
    { .tok = "Intl6",             .code = 0x8c },
    { .tok = "Intl7",             .code = 0x8d },
    { .tok = "Intl8",             .code = 0x8e },
    { .tok = "Intl9",             .code = 0x8f },
    { .tok = "LANG1",             .code = 0x90 },
    { .tok = "LANG2",             .code = 0x91 },
    { .tok = "LANG3",             .code = 0x92 },
    { .tok = "LANG4",             .code = 0x93 },
    { .tok = "LANG5",             .code = 0x94 },
    { .tok = "LANG6",             .code = 0x95 },
    { .tok = "LANG7",             .code = 0x96 },
    { .tok = "LANG8",             .code = 0x97 },
    { .tok = "LANG9",             .code = 0x98 },
    { .tok = "AltErase",          .code = 0x99 },
    { .tok = "SysReq/Attn",       .code = 0x9a },
    { .tok = "Cancel",            .code = 0x9b },
    { .tok = "Clear",             .code = 0x9c },
    { .tok = "Prior",             .code = 0x9d },
    { .tok = "Return",            .code = 0x9e },
    { .tok = "Separator",         .code = 0x9f },
    { .tok = "Out",               .code = 0xA0 },
    { .tok = "Oper",              .code = 0xA1 },
    { .tok = "Clear/Again",       .code = 0xA2 },
    { .tok = "CrSel/Props",       .code = 0xA3 },
    { .tok = "ExSel",             .code = 0xA4 },
    { .tok = "LeftCtrl",          .code = 0xE0 },
    { .tok = "LeftShift",         .code = 0xE1 },
    { .tok = "LeftAlt",           .code = 0xE2 },
    { .tok = "LeftGUI",           .code = 0xE3 },
    { .tok = "RightCtrl",         .code = 0xE4 },
    { .tok = "RightShift",        .code = 0xE5 },
    { .tok = "RightAlt",          .code = 0xE6 },
    { .tok = "RightGUI",          .code = 0xE7 },
};

/*
 *
 */
void matchScanCode(const char *tok, int col, int row)
{
    for (int i = 0; i < (sizeof(scanCodes) / sizeof(scanCodes[0])); i++)
    {
        if (strcmp(tok, scanCodes[i].tok) == 0)
        {
            matrix[col][row] = scanCodes[i].code;
            return;
        }
    }

    fprintf(stderr, "error parsing scancode token: couldn't find \"%s\" on "
            "column %d, row %d\n", tok, col, row);
    exit(1);
}

/*
 *
 */
void readInput(void)
{
    char tok[BUFSIZ];
    for (int row = 0; row < KBD_ROWS; row++)
    {
        for (int col = 0; col < KBD_COLS; col++)
        {
            int matched = fscanf(inputFile, "%s", tok);
            if (matched == EOF)
            {
                if (ferror(inputFile))
                    fprintf(stderr, "error reading input: %s\n",
                            strerror(errno));
                else
                    fprintf(stderr, "error reading input: unexpected EOF\n");

                exit(1);
            }

            matchScanCode(tok, col, row);
        }
    }
}

/*
 *
 */
void writeOutput(void)
{
    unsigned short addr = 0;
    unsigned char checkSum;
    for (int col = 0; col < KBD_COLS; col++)
    {
        checkSum = 0x04 + (addr >> 8) + (addr & 0xff) + 0x00;
        fprintf(outputFile, ":04%04hX00", addr);
        addr += 4;

        for (int row = 0; row < KBD_ROWS; row++)
        {
            fprintf(outputFile, "%02X", matrix[col][row]);
            checkSum += matrix[col][row];
        }
        checkSum = (0x100 - checkSum);
        fprintf(outputFile, "%02X\n", checkSum);
    }
    fputs(":00000001FF\n", outputFile);
}

/*
 *
 */
void showScanCodes(void)
{
    printf("known scancodes:\n");
    for (int i = 0; i < (sizeof(scanCodes) / sizeof(scanCodes[0])); i++)
        printf("  %s\n", scanCodes[i].tok);
    printf("\nthe following keycodes are special:\n");
    printf("  %s:\tkey will be ignored entirely\n", scanCodes[0].tok);
    printf("  %s:\tkey is always pressed, used for calibration\n",
            scanCodes[1].tok);
    printf("  %s:\tkey is never pressed, used for calibration\n",
            scanCodes[2].tok);
}

/*
 *
 */
void showUsage(void)
{
    printf("usage: mtxrom [-h] [-l] [-i inputfile] [-o outputfile]\n");
    printf("  -h\tdisplay this help\n");
    printf("  -l\tlist valid scancode strings\n");
    printf("  -i\tinput file to read human-readable matrix from, defaults to "
            "stdin\n");
    printf("  -o\toutput file to write eeprom ihex matrix file to, defaults to "
            "stdout\n");
}

/*
 *
 */
int main(int argc, char **argv)
{
    inputFile = stdin;
    outputFile = stdout;

    int opt;
    while ((opt = getopt(argc, argv, "h?l")) != -1)
    {
        switch (opt)
        {
            case 'h':
            case '?':
                hflag++;
                break;
            case 'l':
                lflag++;
                break;
        }
    }

    if (hflag)
        showUsage();
    else if (lflag)
        showScanCodes();
    else
    {
        readInput();
        writeOutput();
    }

    return 0;
}
