/******************************************************************************
  Copyright 2013 Tom Cornall

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.  
 ******************************************************************************/
#include <cstring>
#include <errno.h>
#include <iomanip>
#include <iostream>
#include <libusb.h>
#include <stdexcept>
#include <unistd.h>
#include "diag_types.h"

#define DIAG_IFACE_NUM 1
#define DIAG_IFACE_IN_EP 0x83

unsigned char currState[8];

using namespace std;

/*
 *
 */
libusb_device_handle * openDevice(void)
{
    int err;

    libusb_device_handle *dev =
        libusb_open_device_with_vid_pid(NULL, 0x0481, 0x0002);
    if (dev == NULL)
        throw runtime_error("error: could not open device with vid=0x0481 "
                "pid=0x0002");

    err = libusb_kernel_driver_active(dev, DIAG_IFACE_NUM);
    if (err == 1)
    {
        if ((err = libusb_detach_kernel_driver(dev, DIAG_IFACE_NUM)) != 0)
            throw runtime_error(string("error: could not detach kernel driver "
                        "from interface: ") + libusb_error_name(err));
    }
    else if (err != 0)
        throw runtime_error(string("error: could not check if kernel driver "
                    "active: ") + libusb_error_name(err));

    if ((err = libusb_claim_interface(dev, DIAG_IFACE_NUM)) != 0)
        throw runtime_error(string("error: could not claim diag interface: ") +
                libusb_error_name(err));

    return dev;
}

/*
 *
 */
void closeDevice(libusb_device_handle *dev)
{
    int err;

    if ((err = libusb_release_interface(dev, DIAG_IFACE_NUM)) != 0)
        cerr << "warning: had error whilst releasing diag interface: " <<
            libusb_error_name(err) << endl;
    libusb_close(dev);
}

/*
 *
 */
void updateState(libusb_device_handle *dev)
{
    int err, bytesTxd;
    unsigned char buf[8];

    memset(currState, 0x00, sizeof(currState));

    if ((err = libusb_interrupt_transfer(dev, DIAG_IFACE_IN_EP, buf,
                    sizeof(buf), &bytesTxd, 1000)) != 0)
        throw runtime_error(string("error during interrupt transfer: ") +
                libusb_error_name(err));

    memcpy(currState, buf, bytesTxd);
}

/*
 *
 */
void sendCtrl(libusb_device_handle *dev, unsigned char buf[8])
{
    int err;

    if ((err = libusb_control_transfer(dev, LIBUSB_REQUEST_TYPE_CLASS |
                    LIBUSB_RECIPIENT_INTERFACE | LIBUSB_ENDPOINT_OUT, 0x09,
                    (2 << 8) | 0, 1, buf, sizeof(buf), 1000)) != sizeof(buf))
        throw runtime_error(string("error during control transfer: ") +
                libusb_error_name(err));
}

/*
 *
 */
void sendCmd(libusb_device_handle *dev, DiagReportCommand cmd,
        unsigned int data)
{
    unsigned char buf[8];
    buf[0] = cmd;
    buf[1] = data & 0xff;
    buf[2] = data >> 8;
    buf[3] = data >> 16;
    buf[4] = data >> 24;
    sendCtrl(dev, buf);
}

/*
 *
 */
void sendCmd(libusb_device_handle *dev, DiagReportCommand cmd,
        unsigned char p1, unsigned char p2, unsigned char p3)
{
    unsigned char buf[8];
    buf[0] = cmd;
    buf[1] = p1;
    buf[2] = p2;
    buf[3] = p3;
    sendCtrl(dev, buf);
}

/*
 *
 */
void setState(libusb_device_handle *dev, DiagReportState state,
        bool block = true)
{
    sendCmd(dev, dcSetState, state);

    if (block)
    {
        do
        {
            updateState(dev);
        } while (currState[0] != state);
    }
}

/*
 *
 */
void showUsage(void)
{
    cout << "usage: beamspringusb_diag_util [-h] [-r] [-s vref] [-p] [-a] [-m] "
        "[-b] [-x]" << endl;
    cout << "                               [-c CxR=SC] [-L] [-S]" << endl;
    cout << endl;
    cout << "  -h\tdisplay this help" << endl;
    cout << "  -r\tread Vref" << endl;
    cout << "  -s\tset Vref" << endl;
    cout << "  -p\tsave current Vref to EEPROM (first set value to 65535 for "
        << endl;
    cout << "    \tauto-calibration on boot)" << endl;
    cout << "  -a\trun auto-calibration routine" << endl;
    cout << "  -m\tmonitor key states" << endl;
    cout << "  -b\tenter bootloader" << endl;
    cout << "  -x\tshow current matrix" << endl;
    cout << "  -c\tset scancode at (C)olumn and (R)ow to (SC)ancode" << endl;
    cout << "  -L\tload scancodes from EEPROM" << endl;
    cout << "  -S\tsave scancodes to EEPROM" << endl;
}

int hflag;
int rflag;
const char *vrefStr;
int pflag;
int aflag;
int mflag;
int bflag;
int xflag;
const char *scancodeStr;
int Lflag;
int Sflag;

/*
 *
 */
int main (int argc, char **argv)
{
    int err;
    libusb_device_handle *dev = NULL;

    /* read opts */
    int opt;
    while ((opt = getopt(argc, argv, "h?rs:pambxc:LS")) != -1)
    {
        switch (opt)
        {
            case 'r':
                rflag++;
                break;
            case 's':
                vrefStr = optarg;
                break;
            case 'p':
                pflag++;
                break;
            case 'a':
                aflag++;
                break;
            case 'm':
                mflag++;
                break;
            case 'b':
                bflag++;
                break;
            case 'x':
                xflag++;
                break;
            case 'c':
                scancodeStr = optarg;
                break;
            case 'L':
                Lflag++;
                break;
            case 'S':
                Sflag++;
                break;
            case 'h':
            case '?':
                hflag++;
                break;
        }
    }

    if (argc < 2 || hflag)
    {
        showUsage();
        return 0;
    }

    if ((err = libusb_init(NULL)) != 0)
    {
        cerr << "error initialising libusb: " << libusb_error_name(err) << endl;
        return 1;
    }
    libusb_set_debug(NULL, 3);

    try
    {
        dev = openDevice();

        setState(dev, drNothing, false);

        if (rflag)
        {
            setState(dev, drVref);
            cout << (currState[1] + (currState[2] << 8)) << endl;
        }

        if (vrefStr)
        {
            int vref = atoi(vrefStr);
            sendCmd(dev, dcSetVref, vref);
        }

        if (pflag)
            sendCmd(dev, dcStoreVref, 1);

        if (aflag)
        {
            cerr << "Release all keys! Calibration starts in 1 second..." <<
                endl;
            sleep(1);
            cerr << "Running auto-calibration..." << endl;
            sendCmd(dev, dcAutoCalibrate, 1);
            sleep(1);
            setState(dev, drNothing, false);
            setState(dev, drVref);
            cout << "...done, vref: " << (currState[1] + (currState[2] << 8)) <<
                endl;
        }

        if (mflag)
        {
            unsigned char cols[24];

            setState(dev, dr1stCols);
            while (true)
            {
                if (currState[0] == dr1stCols)
                {
                    for (int i = 1, col = 0; i < 7; i++)
                    {
                        cols[col++] = currState[i] & 0xf;
                        cols[col++] = currState[i] >> 4;
                    }
                }
                else if (currState[0] == dr2ndCols)
                {
                    for (int i = 1, col = 12; i < 7; i++)
                    {
                        cols[col++] = currState[i] & 0xf;
                        cols[col++] = currState[i] >> 4;
                    }
                }

                cout << "  |";
                for (int i = 0; i < 23; i++)
                    cout << setw(3) << i;
                cout << endl << "--+";
                for (int i = 0; i < 23; i++)
                    cout << "---";
                cout << endl;
                for (int row = 0; row < 4; row++)
                {
                    cout << row << " |";
                    for (int col = 0; col < 23; col++)
                        cout << "  " << ((cols[col] & (1 << row)) ? '*' : '_');
                    cout << endl;
                }
                cout << endl << endl;

                updateState(dev);
            }
        }

        if (bflag)
            sendCmd(dev, dcJumpBootloader, 1);

        if (xflag)
        {
            unsigned char mtx[23][4];

            setState(dev, drScanCodes);
            for (int i = 0; i < 23; i++)
            {
                for (int i = 0; i < 4; i++)
                    mtx[currState[1]][i] = currState[i + 2];
                updateState(dev);
            }

            cout << "  |";
            for (int i = 0; i < 23; i++)
                cout << setw(3) << i;
            cout << endl << "--+";
            for (int i = 0; i < 23; i++)
                cout << "---";
            cout << endl;
            for (int row = 0; row < 4; row++)
            {
                cout << row << " |";
                for (int col = 0; col < 23; col++)
                    cout << " " << setw(2) << setfill('0') << hex <<
                        (unsigned int)mtx[col][row];
                cout << endl;
            }
        }

        if (scancodeStr)
        {
            unsigned char col, row, scancode;
            int r = sscanf(scancodeStr, "%hhd x %hhd = %hhx", &col, &row,
                    &scancode);
            if (r == 3)
            {
                cerr << "setting scancode at column " << (unsigned int)col <<
                        " and row " << (unsigned int)row << " to 0x" <<
                        hex << (unsigned int)scancode << endl;
                sendCmd(dev, dcSetScanCode, row, col, scancode);
            }
            else
            {
                cerr << "error parsing scancode specifier:" << endl;
                showUsage();
            }
        }

        if (Lflag)
            sendCmd(dev, dcLoadScanCodes, 1);

        if (Sflag)
            sendCmd(dev, dcStoreScanCodes, 1);

        closeDevice(dev);
    }
    catch (exception &e)
    {
        cerr << e.what() << endl;
        if (dev)
            closeDevice(dev);
        return 1;
    }

    return 0;
}
